/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * CharMap.h: Character map and character cell manipilation functions
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "../config.h"

#include <sys/types.h>

#include "global.h"
#include "CharMap.h"

void CharMap::clear(void) {
  POFUNC;
  width=height=0;
  cursx=cursy=0;
  cursv=TRUE;
  cmap=NULL;
}

void CharMap::free(void) {
  POFUNC;
  delete [width*height] cmap;
  clear();
}

BOOL CharMap::alloc(int x, int y) {
  POFUNC;
  width=x; height=y;
  if(!(cmap=new CharCell[x*y]))
    return FALSE;
  return TRUE;
}

CharMap::CharMap(void) {
  POFUNC;
  clear();
}

CharMap::~CharMap() {
  POFUNC;
  free();
}

BOOL CharMap::CompareLines(const CharMap &c2, int line) const {
  // Are the CharMaps valid?
  if(!*this || !c2)
    return TRUE;

  // Are the CharMaps the same size?
  if((width!=c2.width)||(height!=c2.height))
    return TRUE;

  // Check to see if the CharCells are identical in both, if not, fail
  for(int x=0;x<width;x++)
    if(getCharCell(x, line)!=c2.getCharCell(x, line))
      return TRUE;

  // If the flash state has changed, check if there is flashing text on the
  // line
  if(flashcycle!=c2.flashcycle)
    for(int x=0;x<width;x++)
      if(getCharCell(x, line).getFlash())
	return TRUE;
  
  // both lines of map match at this point. Now check for the cursor

  // Is the cursor not on this line in both, if so, success.
  if((cursy!=line)&&(c2.cursy!=line))
    return FALSE;

  // Is cursor invisible in both, if so, success.
  if(((!cursv))&&((!c2.cursv)))
    return FALSE;

  // Now, the cursor is visible in either, and is on this line at least in
  // either. The only way it could possibly not affect the display is if it
  // hasn't moved, and hasn't changed visibility.
  if((cursx==c2.cursx)&&(cursy==c2.cursy)&&(cursv==c2.cursv)&&(curscycle==c2.curscycle))
    return FALSE;

  return TRUE;
}

CharMap &CharMap::operator=(const CharMap &c) {
  POFUNC;

  free();

  if(!&c) { // Assignment from NULL...
    return *this;
  }
  if(setSize(c.width, c.height)) {
    memcpy(cmap, c.cmap, width*height*sizeof(cmap[0]));
  } else {
    // Failed - now what?
  }
  cursx=c.cursx;
  cursy=c.cursy;
  cursv=c.cursv;
  flashcycle=c.flashcycle;
  curscycle=c.curscycle;
  return *this;
}
