/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * GenFont.cc: Utility to convert PBM files containing font data into
 * RawFont binary format.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/time.h>

// Networking stuff
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <iostream.h>
#include <String.h>

// libg++
#include <GetOpt.h>
// undef putc from GetOpt.h
#undef putc

//#include "../config.h"
#include "global.h"
#include "RawFont.h"

#undef PROGNAME
#define PROGNAME "XViewdata-GenFont"

#define DEFAULT_DIR "../fonts/"


/* The header file for libpbm.a doesn't work with C++, so I don't bother
 * including it, and define the functions I need here.
 */

typedef unsigned char bit;
#define PBM_WHITE 0
#define PBM_BLACK 1

extern "C" bit **pbm_readpbm(FILE* file, int* colsP, int* rowsP);
extern "C" FILE *pm_openr(const char *name);
extern "C" void pm_freearray(bit **bits, int rows);
extern "C" void pm_close(FILE* fp);

/*
 * This method reads the pbm encoded font data from the file <name>.
 * It returns NULL for success, or an error otherwise.
 */
int RawFont::LoadPBM(const String &name) {
  POFUNC;

  free();
  D("Reading pbm file " << name);

  int cols, rows;
  FILE *f=pm_openr(name);
  if(!f)
    return errno;

  bit **bits=pbm_readpbm(f, &cols, &rows);
  
  pm_close(f);

  D("Bitmap is " << cols << " by " << rows);
  width=cols/16; height=rows/8;
  D("Font glyphs are " << width << " by " << height);

  int x, y, X, Y;
  int c=0;
  BIT *p;

  D("Converting pbm -> glyphs");
  for(Y=0; Y<8; Y++) {
    for(X=0; X<16; X++) {
      if(!(p=new BIT[width*height])) {
	pm_freearray(bits, rows);
	return -ENOMEM;
      }
      glyph[c]=p;
      for(y=0; y<height; y++) {
	for(x=0; x<width; x++) {
	  *p++=(bits[Y*height+y][X*width+x])?(~0):0;
	}
      }
      c++;
    }
  }

  pm_freearray(bits, rows);

  return 0;
}

void show_options(void) {
  cerr
    << PROGNAME " version " VERSION " (" __DATE__  " " __TIME__ ")\n"
    <<"Copyright (C) 1997 Peter Corlett. All rights reserved.\n"
    << "This software comes with NO WARRANTY: see the file COPYING for details.\n"
    << "\n"
    << "Usage: " PROGNAME " [options] file file ...\n"
    << "\n"
    << "Options:\n"
    << "-o <s>\tOutput raw fonts to directory s (default \"" << DEFAULT_DIR << "\")\n"
    << "file\tPBM format datafile to convert\n"
    << "\n";
}

int main(int argc, char *argv[]) {
  PGFUNC;

  GetOpt getopt(argc, argv, "o:h");
  int option_char;

  String odir=DEFAULT_DIR;

  while((option_char=getopt())!=EOF)
    switch(option_char) {
    case 'o':
      odir=getopt.optarg;
      break;
    case 'h':
    default:
      show_options();
      return 1;
    }

  int n=getopt.optind;

  while(n<argc) {
    RawFont f;
    String fname;
    f.LoadPBM(argv[n]);
    
    // it appears that the String class doesn't know how to convert from
    // integers so I use this kludge.
    char rname[80];
    sprintf(rname, "%dx%d.font", f.getWidth(), f.getHeight());
    fname=odir+"/"+rname;

    /* FIXME - check return code */
    f.SaveBin(fname);
    n++;
  }

  return 0;
}
