/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * Viewdata.cc - Implementation of a Viewdata terminal driver
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../config.h"

#include <unistd.h>

#include "global.h"
#include "TerminalDriver.h"
#include "Viewdata.h"

#include <X11/keysym.h>

#include <iostream.h>

static int ViewdataToASCII(int ch) {
  switch(ch) {
  case '#': return 19;  // '#' -> UK pound symbol
  case '[': return 27;  // '[' -> left arrow
  case '\\': return 17; // '\' -> one half
  case ']': return 28;  // `]` -> right arrow
  case '^': return 26;  // '^' -> up arrow
  case '_': return '#'; // '_' -> '#'
  case '`': return 25;  // '`' -> horizontal bar
  case '{': return 16;  // '{' -> one quarter
  case '|': return 29;  // '|' -> double vertical bar
  case '}': return 17;  // '}' -> three quarters
  case '~': return 30;  // '~' -> divide symbol
  case 127: return 31;  // 127 -> square block
  default: return ch;
  }
}

void Viewdata::cls(void) {
  POFUNC;
  int x,y;
  for(y=cmap.getHeight()-1;y>=0;y--) {
    for(x=cmap.getWidth()-1;x>=0;x--)
      vmap[y*cmap.getWidth()+x]=' ';
  }

}

/* This is the code table used:
   [NU] AR  AG  AY  AB  AM  AC  AW
    FL  ST [EB][SB] NH  DH [SO][SI]
   [DL] GR  GG  GY  GB  GM  GC  GW
    CD  CG  SG [EC] BB  NB  HG  RG
    */

static void TeletextToCharCell(CharMap &cmap, UBYTE *teletext) {
  PGFUNC;
  int x, y;
  int f, b, fl, gfx, sepg, hold;
  int dh, dhline=0;
  UBYTE ch, last;
  CharCell *ccell=&cmap.getCharCell();
  CharCell newc;
  
  for(y=0;y<cmap.getHeight();y++) {
    f=7; b=0; fl=0; gfx=0; sepg=0; dh=0;
    hold=0; last=32;
    for(x=0;x<cmap.getWidth();x++) {
      if(dhline==2) {
	ch=teletext[-cmap.getWidth()]&127;
	teletext++;
      }	else {
	ch=(*teletext++)&127;
      }
      switch(ch) {
      case 0: /* no-op: NUL */
	break;
      case 1 ... 7: /* alpha red...white */
	f=ch; gfx=0; break;
      case 8: /* flash */
	fl=1; break;
      case 9: /* steady */
	fl=0; break;
      case 10 ... 11: /* no-op: end box, start box */
	break;
      case 12: /* normal height */
	dh=0;
	break;
      case 13: /* double height */
	dh=1; if(dhline==0) dhline=1;
	break;
      case 14 ... 16: /* no-op: shift out, shift in, DLE */
	break;
      case 17 ... 23: /* graphics red...white */
	f=ch-16; gfx=1; break;
      case 24: /* conceal display */
	break;
      case 25: /* contiguous graphics */
	sepg=0; break;
      case 26: /* separated graphics */
	sepg=1; break;
      case 27: /* no-op: ESC */
	break;
      case 28: /* black background */
	b=0; break;
      case 29: /* new background */
	b=f; break;
      case 30: /* hold graphics */
	hold=1; break;
      case 31: /* release graphics */
	hold=0; break;
      }
      if(ch<32) {
	if(hold) {
	  ch=last;
	} else {
	  ch=32;
	}
      } else {
	last=ch;
      }
      /* We have the code of the character to display. Now, we either
       * (a) Display it directly (not in dh mode, not on second line)
       * (b) Display it as a graphics char (gfx&&(ch*32))
       * (c) Display the top half of it (in dh mode, on first line)
       * (d) Display the bottom half of it (in dh mode, on second line)
       * (e) Suppress it (not in dh mode, but on second line)
       */
      if(gfx&&(ch&32))
	newc.setGlyph(128|(ch&0x1f)|((ch&0x40)?0x20:0)|(sepg?64:0));
      else {
	/* It's a normal Viewdata text character - convert it to ASCII if
	 * possible - use special glyphs from font otherwise.
	 */
	ch=ViewdataToASCII(ch);
	if(dh) {
	  if(dhline==2)
	    newc.setGlyph(ch+384);
	  else
	    newc.setGlyph(ch+256);
	} else {
	  if(dhline==2)
	    newc.setGlyph(32);
	  else
	    newc.setGlyph(ch);
	}
      }
      newc.setFore(f);
      newc.setBack(b);
      newc.setFlash(fl);
      if(newc!=*ccell) {
	*ccell=newc;
      }
      ccell++;
    }
    if(dhline==1) {
      dhline=2;
    } else if(dhline==2) {
      dhline=0;
    }
  }
}

void Viewdata::clear() {
  POFUNC;
  vmap=NULL;
  escpending=FALSE;
  ctrl_key=FALSE;
}

void Viewdata::free() {
  POFUNC;
  if(vmap)
    delete [cmap.getWidth()*cmap.getHeight()] vmap;
  clear();
}

Viewdata::Viewdata(void) {
  POFUNC;
  clear();

  setSize(40, 24);
}

Viewdata::~Viewdata() {
  POFUNC;
  free();
}

BOOL Viewdata::setSize(int x, int y) {
  POFUNC;
  free();
  if(!(vmap=new UBYTE [x*y]))
    return FALSE;
  if(!cmap.setSize(x, y))
    return FALSE;
  cls();

  D("Okay, resized to " << x << "x" << y);
  return TRUE;
}

int Viewdata::putc(UBYTE c) {
  POFUNC;
  if(!vmap) {
    D("Warning: no Viewdata character map to write to!");
    return TRUE;
  }

  if(escpending) {
    escpending=FALSE;
    c+=64;
  }

  switch(c) {
  case 7: /* Bell */
    /* I wonder what I should do here...? */
    break;
  case 8: /* Cursor left */
    cmap.setCursorX(cmap.getCursorX()-1);
    if(cmap.getCursorX()>=0)
      break;
    cmap.setCursorX(cmap.getWidth()-1);
    /* Fallthru */
  case 11: /* Cursor up */
    cmap.setCursorY(cmap.getCursorY()-1);
    if(cmap.getCursorY()<0)
      cmap.setCursorY(cmap.getHeight()-1);
    break;
  case 32 ... 255: /* Printable character */
    vmap[cmap.getCursorX()+cmap.getCursorY()*cmap.getWidth()]=c;
    /* Fallthru */
  case 9: /* Cursor right */
    cmap.setCursorX(cmap.getCursorX()+1);
    if(cmap.getCursorX()<cmap.getWidth())
      break;
    cmap.setCursorX(0);
    /* Fallthru */
  case 10: /* Cursor down */
    cmap.setCursorY(cmap.getCursorY()+1);
    if(cmap.getCursorY()>=cmap.getHeight()) {
      #if 1
      cmap.setCursorY(0);
      #else
      cmap.setCursorY(cmap.getHeight()-1);
      for(int x=0; x<cmap.getWidth(); x++) {
	for(int y=1; y<cmap.getHeight(); y++) {
	  vmap[x+(y-1)*cmap.getWidth()]=vmap[x+y*cmap.getWidth()];
	}
	vmap[x+(cmap.getHeight()-1)*cmap.getWidth()]=32;
      }
      #endif
    }
    break;
  case 12: /* Clear Screen */
    cls();
    cmap.setCursorX(0); cmap.setCursorY(0);
    break;
  case 13: /* CR / Start of line */
    cmap.setCursorX(0);
    break;
  case 17: /* Cursor on */
    cmap.setCursorVisibility(TRUE);
    break;
  case 20: /* Cursor off */
    cmap.setCursorVisibility(FALSE);
    break;
  case 27: /* Escape */
    escpending=TRUE;
    break;
  case 30: /* Cursor home (0,0) */
    cmap.setCursorX(0);
    cmap.setCursorY(0);
    break;
  case 31: /* Gareth Babb's Wizzo code - go to bottom of screen */
    cmap.setCursorX(0);
    cmap.setCursorY(cmap.getHeight()-1);
  default: /* Some other character */
    break;
  }
  return FALSE;
}

int Viewdata::write(const char *buf, int len) {
  POFUNC;

  int err;
  for(int n=0;n<len;n++) {
    if((err=putc(buf[n])))
      return err;
  }

  TeletextToCharCell(cmap, vmap);
  return FALSE;
}

int Viewdata::keyupread(char *buf, int len, int k) {
  POFUNC;

  switch(k) {
  case XK_Control_L:
  case XK_Control_R:
    ctrl_key=FALSE;
    D("Control key up");
    break;
  }

  return 0;
}

int Viewdata::read(char *buf, int len, int k) {
  POFUNC;
  char c;

  if(len<2)
    return -1;

  switch(k) {
  case XK_Control_L:
  case XK_Control_R:
    ctrl_key=TRUE;
    D("Control key down");
    return 0;
    break;

  case XK_Left:
  case XK_BackSpace:
  case XK_Delete:
    c=8; break;

  case XK_Tab:
  case XK_Right:
    c=9; break;

  case XK_Down:
    c=10; break;

  case XK_Up:
    c=11; break;

  case XK_Insert:
    c=12; break;

  case XK_Return:
    c=13; break;

  case XK_Page_Up:
    c=17; break;

  case XK_Page_Down:
    c=20; break;

  case XK_Escape:
    c=27; break;

  case XK_Home:
    c=30; break;

  case XK_End:
    c=31; break;

  case 0xa3: /* Convert '' to 35 */
  case 96: /* Convert '`' to 35 (for non-UK keyboards) */
    c=35; break;
  case 35: /* Convert '#' to 95 */
    c=95; break;
  case 95: /* Convert '_' to 96 */
    c=96; break;

  case 32 ... 34:
  case 36 ... 94:
  case 97 ... 126:
    c=k; break;
    
  case XK_KP_0 ... XK_KP_9: /* Numeric keypad 0-9 */
    c=k-XK_KP_0+'0'; break;

  case XK_KP_Enter:
    c=13; break;

  case XK_KP_Equal:
    c='='; break;

  case XK_KP_Multiply:
    c='*'; break;

  case XK_KP_Add:
    c='+'; break;

  case XK_KP_Subtract:
    c='-'; break;

  case XK_KP_Decimal:
    c='.'; break;

  case XK_KP_Divide:
    c='/'; break;

  default:
    // cerr << "Unknown keysym " << hex << k << dec << "\n";
    return 0;
  }

  if(ctrl_key)
    if(c=='.') /* kludge to xlate C-. to 127 */
      c=127;
    else
      c&=31;

  buf[0]=c;
  return 1;
}

Viewdata *newViewdata(const String &arg) {
  PGFUNC;
  Viewdata *v;

  if(!(v=new Viewdata))
    return NULL;

  return v;
}
